use std::{
    cmp::min,
    iter::zip,
    ops::{Add, AddAssign},
};

use super::{Number, Sign};

impl Add for Number {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        let self_len = self.digits.len();
        let rhs_len = rhs.digits.len();
        let mut self_digits = self.digits.clone();
        let mut rhs_digits = rhs.digits.clone();
        if self_len != rhs_len {
            let difference = (self_len).abs_diff(rhs_len);
            let pad = vec![0i8; difference];
            if min(self_len, rhs_len) == self_len {
                self_digits = [self.digits, pad].concat();
            } else {
                rhs_digits = [rhs.digits, pad].concat();
            }
        }
        let zipped = zip(self_digits.iter(), rhs_digits.iter());
        let mut carry = 0;
        let mut digits = Vec::new();
        let mut sign = match (self.sign, rhs.sign) {
            (Sign::Positif, Sign::Positif) => Sign::Positif,
            (Sign::Negatif, Sign::Negatif) => Sign::Negatif,
            (Sign::Positif, Sign::Negatif) | (Sign::Negatif, Sign::Positif) => Sign::Positif,
        };
        for (a, b) in zipped {
            let comb = match (self.sign, rhs.sign) {
                (Sign::Positif, Sign::Positif) => {
                    let comb = a + b + carry;
                    if comb > 9 {
                        carry = 1;
                    } else {
                        carry = 0;
                    }
                    comb % 10
                }
                (Sign::Positif, Sign::Negatif) => {
                    let comb = a - b - carry;
                    if comb > 0 && b > a {
                        carry = 1;
                    } else {
                        carry = 0;
                    }
                    if b > a {
                        sign = Sign::Negatif;
                    } else {
                        sign = Sign::Positif;
                    }
                    (comb % 10).abs()
                }
                (Sign::Negatif, Sign::Positif) => {
                    let comb = -a + b - carry;
                    if comb > 0 && a > b {
                        carry = 1;
                    } else {
                        carry = 0;
                    }
                    if a > b {
                        sign = Sign::Negatif;
                    } else {
                        sign = Sign::Positif;
                    }
                    (comb % 10).abs()
                }
                (Sign::Negatif, Sign::Negatif) => {
                    let comb = a + b + carry;
                    if comb > 9 {
                        carry = 1;
                    } else {
                        carry = 0;
                    }
                    comb % 10
                }
            };
            digits.push(comb);
        }

        if carry != 0 {
            digits.push(carry);
        }

        Self { digits, sign }
    }
}

impl AddAssign for Number {
    fn add_assign(&mut self, rhs: Self) {
        let new = self.clone() + rhs;
        self.digits = new.digits;
        self.sign = new.sign;
    }
}

#[cfg(test)]
mod test_number_add {
    use crate::number::Number;

    #[test]
    fn add_positif() {
        let a = Number::from(1);
        let b = 1.into();
        let res = Number::from(2);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_large_positif() {
        let a = Number::from(9);
        let b = 9.into();
        let res = Number::from(18);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_negatif() {
        let a = Number::from(-1);
        let b = (-1).into();
        let res = Number::from(-2);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_large_negatif() {
        let a = Number::from(-9);
        let b = (-9).into();
        let res = Number::from(-18);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_negatif_to_positif() {
        let a = Number::from(1);
        let b = (-1).into();
        let res = Number::from(0);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_positif_to_negatif() {
        let a = Number::from(-1);
        let b = (1).into();
        let res = Number::from(0);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_larger_positif_to_negatif() {
        let a = Number::from(-1);
        let b = (11).into();
        let res = Number::from(10);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_larger_negatif_to_positif() {
        let a = Number::from(1);
        let b = (-11).into();
        let res = Number::from(-10);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_positif_to_larger_negatif() {
        let a = Number::from(-99);
        let b = (11).into();
        let res = Number::from(-88);
        assert_eq!(res, a + b);
    }

    #[test]
    fn add_negatif_to_larger_positif() {
        let a = Number::from(99);
        let b = (-11).into();
        let res = Number::from(88);
        assert_eq!(res, a + b);
    }
}
