use core::panic;
use std::{
    error::Error,
    ops::{Div, DivAssign, Rem, RemAssign},
};

use super::{Number, Sign};

impl Number {
    pub fn div_with_rem(n1: Number, n2: Number) -> Result<(Number, Number), Box<dyn Error>> {
        if n2 == 0.into() {
            return Err(Box::from("Division by 0"));
        }
        let n1_len = n1.digits.len();
        let n2_len = n2.digits.len();
        if n2_len > n1_len {
            return Ok((Number::from(0), n2));
        }
        let dividend = n1.digits[..n2_len].to_vec();
        let mut quotient = vec![];
        let mut remainder = Number {
            digits: dividend.clone(),
            sign: Sign::Positif,
        };
        let mut iteration = 1;
        loop {
            let mut factor = 0;
            loop {
                let temp_remainder = remainder.clone() - n2.clone();
                if temp_remainder.sign == Sign::Negatif {
                    quotient.push(factor);
                    break;
                }
                remainder = temp_remainder;
                factor += 1;
            }
            if n1_len == n2_len + iteration - 1 {
                break;
            }
            remainder.digits.push(n1.digits[n2_len + iteration - 1]);
            iteration += 1;
        }
        let mut res = Number {
            digits: quotient,
            sign: Sign::Positif,
        };
        for digit in res.clone().digits {
            if digit != 0 || res.digits.len() == 1 {
                break;
            }
            res.digits.remove(0);
        }
        for digit in remainder.clone().digits {
            if digit != 0 || remainder.digits.len() == 1 {
                break;
            }
            remainder.digits.remove(0);
        }
        Ok((res, remainder))
        //if n2 == 0.into() {
        //    return Err(Box::from("Division by 0"));
        //}
        //let n1_len = n1.digits.len();
        //let n2_len = n2.digits.len();
        //if n2_len > n1_len {
        //    return Ok((Number::from(0), n1));
        //}
        //let sign = match (n1.sign, n2.sign) {
        //    (Sign::Positif, Sign::Positif) | (Sign::Negatif, Sign::Negatif) => Sign::Positif,
        //    (Sign::Positif, Sign::Negatif) | (Sign::Negatif, Sign::Positif) => Sign::Negatif,
        //};
        //let mut res = Number::from(0);
        //let mut remainder = n1;

        //loop {}

        //if res.digits != vec![0] {
        //    res.sign = sign;
        //}
        //Ok((res, remainder))
    }

    pub fn gcd(mut a: Number, mut b: Number) -> Number {
        loop {
            let t = b.clone();
            b = a % b;
            a = t;
            if b == Number::from(0) {
                return a;
            }
        }
    }
}

impl Div for Number {
    type Output = Self;

    fn div(self, rhs: Self) -> Self::Output {
        Self::div_with_rem(self, rhs).unwrap().0
    }
}

impl DivAssign for Number {
    fn div_assign(&mut self, rhs: Self) {
        match Self::div_with_rem(self.clone(), rhs) {
            Ok((new, _)) => {
                self.digits = new.digits;
                self.sign = new.sign;
            }
            Err(e) => panic!("{e}"),
        }
    }
}

impl Rem for Number {
    type Output = Self;

    fn rem(self, rhs: Self) -> Self::Output {
        Self::div_with_rem(self, rhs).unwrap().1
    }
}

impl RemAssign for Number {
    fn rem_assign(&mut self, rhs: Self) {
        match Self::div_with_rem(self.clone(), rhs) {
            Ok((_, new)) => {
                self.digits = new.digits;
                self.sign = new.sign;
            }
            Err(e) => panic!("{e}"),
        }
    }
}

#[cfg(test)]
mod test_number_div_rem {
    use core::panic;

    use crate::number::Number;

    #[test]
    fn zero_denominator() {
        let a = Number::from(9);
        let b = Number::from(0);
        if Number::div_with_rem(a, b).is_ok() {
            panic!("Division by 0 not allowed");
        }
    }

    #[test]
    fn div_rem_positif() {
        let a = Number::from(9);
        let b = Number::from(9);
        let res = (Number::from(1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_negatif() {
        let a = Number::from(-9);
        let b = Number::from(-9);
        let res = (Number::from(1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_positif_with_negatif() {
        let a = Number::from(9);
        let b = Number::from(-9);
        let res = (Number::from(-1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_negatif_with_positif() {
        let a = Number::from(-9);
        let b = Number::from(9);
        let res = (Number::from(-1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_eleven_x_ten() {
        let a = Number::from(11);
        let b = Number::from(10);
        let res = (Number::from(1), Number::from(1));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_eleven_x_eleven() {
        let a = Number::from(10);
        let b = Number::from(11);
        let res = (Number::from(0), Number::from(10));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_big_1() {
        //let a = Number::from("123");
        //let b = Number::from("32");
        //let res = (Number::from(3), Number::from(27));
        //assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_big_2() {
        //let a = Number::from("123456789");
        //let b = Number::from("123");
        //let res = (Number::from(1003713), Number::from(90));
        //assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }
}
